# This file configures Charmcraft.
# See https://juju.is/docs/sdk/charmcraft-config for guidance.

name: {{ name }}

type: charm

bases:
  - build-on:
    - name: ubuntu
      channel: "22.04"
    run-on:
    - name: ubuntu
      channel: "22.04"

# (Required)
summary: A very short one-line summary of the Flask application.

# (Required)
description: |
  A comprehensive overview of your Flask application.

extensions:
  - flask-framework

# Uncomment the integrations used by your application
# Integrations set to "optional: false" will block the charm
# until the applications are integrated.
# requires:
#   mysql:
#     interface: mysql_client
#     optional: false
#     limit: 1
#   postgresql:
#     interface: postgresql_client
#     optional: false
#     limit: 1
#   mongodb:
#     interface: mongodb_client
#     optional: false
#     limit: 1
#   redis:
#     interface: redis
#     optional: false
#     limit: 1
#   s3:
#     interface: s3
#     optional: false
#     limit: 1
#   saml:
#     interface: saml
#     optional: false
#     limit: 1
